/*
 * RefcountDB.h
 *
 *  Created on: Mar 28, 2022
 *      Author: medsrv
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_REFCOUNTDB_H_
#define IMUTILS_INCLUDE_IMUTILS_REFCOUNTDB_H_

#include <string>
#include <list>
#include "imutils/DBConnector.h"

namespace erad::db {


/**
 * Interface to manage reference counters
 */
class RefcountDB {

	public:
		virtual ~RefcountDB() {
		}

		/**
		 * Adds a new entry to the reference counter.
		 * A new entry will be added even if the reference counter already have an entry with the same name.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 * @param entry the entry to be added
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of entries added.
		 */
		virtual size_t addEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false ) = 0;

		/**
		 * Creates a new entry to the reference counter.
		 * If the reference counter already have an entry with the same name, then no new entry will be created.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 * @param entry the entry to be added
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of entries added.
		 */
		virtual size_t createEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false ) = 0;

		/**
		 * Removes a single entry from the reference counter.
		 * If there are multiple entries with the same name on a reference counter, then only one entry will be removed.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 * @param entry the entry to be added
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of entries removed
		 */
		virtual size_t removeEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false ) = 0;

		/**
		 * Checks if the reference counter has any entries
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if and only if the reference counter has at least on entry
		 */
		virtual bool entriesExist( const std::string& id, const std::string& type ) noexcept ( false ) = 0;

		/**
		 * Gets the count of all entries on a reference counter.
		 * Multiple entries (same entry several times) will be counted with their multiplicity.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of all entries on the reference counter
		 */
		virtual size_t countEntries( const std::string& id, const std::string& type ) noexcept ( false ) = 0;

		/**
		 * Gets the count (multiplicity) of a single entry on a reference counter.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 * @param entry the entry (referrer)
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the multiplicity of the entry on the reference counter
		 */
		virtual size_t countEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false ) = 0;

		/**
		 * Lists all entries from a reference counter.
		 * Multiple entries (same entry several times) will be listed multiple times according to their multiplicity.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 * @param entries output parameters for the entries
		 * @param comment output parameter for the reference counter (only legacy reference counter files had a comment)
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual void listEntries( const std::string& id, const std::string& type,
				std::list<std::string>& entries /*out*/, std::string& comment /*out*/) noexcept ( false ) = 0;

		/**
		 * Creates a lock for a reference counter object
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if and only if the lock could be successfully created
		 */
		virtual bool createLockEntryInDB( const std::string& id, const std::string& type ) noexcept ( false ) = 0;

		/**
		 * Removes a lock from the reference counter object
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual void removeLockEntryFromDB( const std::string& id, const std::string& type ) noexcept ( false ) = 0;

		/**
		 * Collects all references for a reference counter, especially all `OBJECTFILE` references for a study.
		 *
		 * @param refs (out) output parameter for the references, (map of object filenames and list of references)
		 * @param id the reference counter ID (expected to be the study instance UID)
		 * @param type the reference counter type (expected to be `OBJECTFILE`)
		 * @param includeIndirect `true` indicates to also include indirect (`database`) references
		 *
		 * @return `true` if collecting all references is supported for this type
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual bool collectAllRefs( std::map<std::string, std::list<std::string>>& refs, const std::string& id,
				const std::string& type, bool includeIndirect ) noexcept( false ) = 0;

		/**
		 * Drops all references from the reference counter without invoking the *on release* handler.
		 *
		 * Also supports dropping all references from all `OBJECTFILE` reference counters in a study.
		 * This requires to set the reference counter *type* to `ALL_OBJECTFILES` (a pseudo-type),
		 * and the reference counter *id* to the study instance UID.
		 *
		 * @param id the reference counter id
		 * @param type the reference counter type
		 *
		 * @return the count of references removed.
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual size_t dropAllRefs( const std::string& id, const std::string& type ) noexcept( false ) = 0;

		/**
		 * Adjusts the count of *file* references on the `STUDYDIR` reference counter of a particular study to
		 * match the expected reference counts.
		 *
		 * The expected reference count is the number of objects
		 * - having an object table entry, or
		 * - having any `OBJECTFILE` reference.
		 *
		 * Callers need to ensure that object table entries are available in the database for the study, that is
		 * need to restore object table entries when swapped out.
		 *
		 * @param study
		 *          the study instance UID
		 * @return the count of affected RefCount table records (1 or 0)
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual size_t adjustFileRefs( const std::string& study ) noexcept( false ) = 0;
};


/**
 * Reference counter implementation managing reference counters in MySQL database tables.
 */
class RefcountDBSql : public RefcountDB, public PSCache {

	public:

		/**
		 * Constructs a new DevicesDBSql object to prepare and execute MySQL queries
		 *
		 * @param pDB connector to the MySQL database
		 */
		RefcountDBSql( DBConnectorSql& pDB );
		virtual ~RefcountDBSql();

		/**
		 * @copydoc RefcountDB::addEntry(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t addEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false )
				override;

		/**
		 * @copydoc RefcountDB::createEntry(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t createEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false )
				override;

		/**
		 * @copydoc RefcountDB::createEntry(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t removeEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false )
				override;

		/**
		 * @copydoc RefcountDB::entriesExist(const std::string&,const std::string&)
		 */
		virtual bool entriesExist( const std::string& id, const std::string& type ) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::countEntries(const std::string&,const std::string&)
		 */
		virtual size_t countEntries( const std::string& id, const std::string& type ) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::countEntry(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t countEntry( const std::string& id, const std::string& type, const std::string& entry ) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::listEntries(const std::string&,const std::string&,std::list<std::string>&,std::string&)
		 */
		virtual void listEntries( const std::string& id, const std::string& type,
				std::list<std::string>& referrers /*out*/, std::string& comment /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::createLockEntryInDB(const std::string&,const std::string&)
		 */
		virtual bool createLockEntryInDB( const std::string& id, const std::string& type ) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::removeLockEntryFromDB(const std::string&,const std::string&)
		 */
		virtual void removeLockEntryFromDB( const std::string& id, const std::string& type ) noexcept ( false ) override;

		/**
		 * @copydoc RefcountDB::collectAllRefs(std::map<std::string,std::list<std::string>>& const std::string&,const std::string&,bool)
		 */
		virtual bool collectAllRefs( std::map<std::string, std::list<std::string>>& refs, const std::string& id,
				const std::string& type, bool includeIndirect ) noexcept( false ) override;

		/**
		 * @copydoc RefcountDB::dropAllRefs(const std::string&,const std::string&)
		 */
		virtual size_t dropAllRefs( const std::string& id, const std::string& type ) noexcept( false ) override;

		/**
		 * @copydoc RefcountDB::adjustFileRefs(const std::string&)
		 */
		virtual size_t adjustFileRefs( const std::string& study ) noexcept( false ) override;

	private:

		/**
		 * Checks if an object exists in the database
		 *
		 * @param id the object id in 'study/fname' format.
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true id and only if the object exists in the database
		 */
		bool isObjectFileInDatabase( const std::string& id ) noexcept ( false );

		/**
		 * Removes an object from the database
		 *
		 * @param id the object id in 'study/fname' format.
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the count of object table records removed
		 */
		size_t removeObjectFromDatabase( const std::string& id ) noexcept ( false );

		/**
		 * Extracts study UID and file name from an object id
		 *
		 * @param id the object id in 'study/fname' format.
		 * @param study output parameter for the study UID
		 * @param fname output parameter for the file name
		 *
		 * @throw DBException if the object id cannot be parsed
		 *
		 * @return the count of object table records removed
		 */
		void parseObjectId( const std::string& id, std::string& study /* out */, std::string& fname /* fname */) noexcept ( false );


		/**
		 * Collects all references from reference counters matching to a particular pattern.
		 * For example to collect all references from `OBJECTFILE` reference counters from a study.
		 *
		 * @param idPattern      pattern for the reference counter IDs using `SQL LIKE` syntax. ('<study>/%')
		 * @param patternLength  well, the length of the study uid (assumes the pattern is a kind of prefix)
		 * @param type the reference counter type
		 * @param escape escape character for the SQL query (to avoid a hard coded escape character conflict the pattern)
		 * @param refs (out) output parameter for the collected references (map of ids and reference lists)
		 *
		 * @throw DBException if the object id cannot be parsed
		 */
		void collectAllReferences( const std::string& idPattern, size_t patternLen, const std::string& type,
				const std::string& escape, std::map<std::string, std::list<std::string>>& refs ) noexcept( false );

		/**
		 * Collects all (indirect) `database` references for `OBJECTFILE` reference counters for  particular study.
		 *
		 * @param study the study instance UID
		 * @param refs (out) output parameter for the collected references (map of ids and reference lists)
		 *
		 * @throw DBException if the object id cannot be parsed
		 */
		void collectObjectDbRefs( const std::string& study, std::map<std::string, std::list<std::string>>& refs ) noexcept( false );

		/**
		 * Deletes all object table entries for a particular study.
		 *
		 * To be used for dropping all `OBJECTFILE` references without invoking the *on release* handler.
		 *
		 * @param study the study instance UID
		 *
		 * @return the count of the objects deleted
		 *
		 * @throw DBException if the object id cannot be parsed
		 */
		size_t deleteAllObjects( const std::string& study ) noexcept( false );

		/**
		 * Deletes all references from a reference counter.
		 * To be used to drop all references without invoking the *on release* handler.
		 *
		 * @param id the reference counter ID
		 * @param type the reference counter type
		 *
		 * @return the count of the references deleted
		 *
		 * @throw DBException if the object id cannot be parsed
		 */
		size_t deleteReferences( const std::string& id, const std::string& type ) noexcept( false );

		/**
		 * Deletes all references from all reference counters matching to a particular pattern.
		 * For example to delete all references from all `OBJECTFILE` reference counters from a study.
		 *
		 * @param idPattern      pattern for the reference counter IDs using `SQL LIKE` syntax. ('<study>/%')
		 * @param type the reference counter type
		 * @param escape escape character for the SQL query (to avoid a hard coded escape character conflict the pattern)
		 *
		 * @return the count of the references deleted
		 */
		size_t deleteAllReferences( const std::string& idPattern, const std::string& type, const std::string& escape ) noexcept(
				false );

		std::string refCountTbl;
		std::string objectTbl;

		PsQuery addQuery; // = "INSERT INTO RefCounter ( ID, TYPE, REFERRER, COUNTER ) values (?,?,?,1) ON DUPLICATE KEY UPDATE COUNTER=COUNTER+1;";
		PsQuery createQuery; // = "INSERT IGNORE INTO RefCounter ( ID, TYPE, REFERRER, COUNTER ) values (?,?,?,1);";
		PsQuery decrementQuery; // = "UPDATE RefCounter set COUNTER=COUNTER-1 where ID=? and TYPE=? and REFERRER=?;";
		PsQuery removeQuery; // = "DELETE FROM RefCounter where ID=? and TYPE=? and REFERRER=? and COUNTER<1;";
		PsQuery countQuery; // = "select sum(COUNTER) as TOTAL from RefCounter where COUNTER > 0 and ID=? and TYPE=?";
		PsQuery countRefQuery; // = "select COUNTER from RefCounter where ID=? and TYPE=? AND REFERRER=?";
		PsQuery existsQuery; // = "SELECT ID FROM RefCounter WHERE ID=? and TYPE=? and COUNTER>0 LIMIT 1";
		PsQuery checkObjectTableQuery; //= "SELECT FNAME FROM Dcobject WHERE FNAME=? AND STYIUID=?";
		PsQuery removeObjectTableQuery; // = "DELETE FROM Dcobject WHERE FNAME=? AND STYUID=?";
		PsQuery insertLockQuery; // = "INSERT INTO RefCounter ( ID, TYPE, REFERRER, COUNTER ) values (?,?,'LOCKREMOVEREF',0)";
		PsQuery removeLockQuery; // = "DELETE FROM RefCounter WHERE ID=? and TYPE=? and REFERRER='LOCKREMOVEREF'";
		PsQuery listReferrersQuery; // = "SELECT REFERRER, COUNTER from RefCounter where ID=? and TYPE=?";
		PsQuery collectAllRefsQuery;	// = "SELECT ID, REFERRER, COUNTER FROM RefCounter WHERE TYPE=? AND ID LIKE ? ESCAPE ?";
		PsQuery studyObjectsQuery;		// = "SELECT FNAME FROM Dcobject WHERE STYIUID=?";
		PsQuery deleteRefsQuery;		// = "DELETE FROM RefCounter WHERE TYPE=? AND ID=?";
		PsQuery deleteAllRefsQuery;	// = "DELETE FROM RefCounter WHERE TYPE=? AND ID LIKE ? ESCAPE ?";
		PsQuery deleteAllObjsQuery;	// = "DELETE FROM Dcobject WHERE STYIUID=?";
		PsQuery adjustFileRefsQuery;
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_REFCOUNTDB_H_ */
