#ifndef ERAD_UTILS_QUICKDELETER_H_
#define ERAD_UTILS_QUICKDELETER_H_

#include <set>
#include <map>
#include <string>
#include <memory>

#include <imutils/Config.h>
#include <imutils/DBConnector.h>

namespace erad::utils {

/**
 * Status values for `isDeletable()`.
 * Provides enum values about what would remain after a study delete.
 */
enum DeletableStatus {
	/// study directory with objects would be deleted (none to remain)
	None = 0,
	/// study objects would be deleted, but study directory would remain
	Directory = 1,
	/// study objects would remain with forward to external archive references only.
	ArchiveFwd = 2,
	/// study objects would remain
	Objects = 3
};


class ArchiveConfigReader;

/**
 * Archive configuration, especially for external archives.
 * To be used by `isDeletable()` checks.
 */
class ArchiveConfig {
		friend class ArchiveConfigReader;
	private:
		/// The name of the external archive
		std::string m_external;
		/// The name of the `forward` references indicating pending forwards to this device
		std::string m_forwardReference;
		/// The name of the `commit` references indicating pending commits from this device
		std::string m_commitReference;

	protected:

		/**
		 * Constructs a new `ArchiveConfig` object with the provided external archive.
		 * The reference names are generated from the archive name.
		 *
		 * @param external the configured external archive
		 */
		ArchiveConfig( const std::string& external );

	public:

		~ArchiveConfig();

		/**
		 * Checks if an external archive is configured.
		 *
		 * @return `true` if and only if an external archive is configured.
		 */
		bool isExternal() const
		{
			return !m_external.empty();
		}

		/**
		 * Gets the name of the *forward* reference indicting pending forwards toward the external archive.
		 * Will return an empty string if there is no external archive configured (`isExternal()` returns`false`).
		 *
		 * @return the name of the *forward* reference.
		 */
		const std::string& getForwardReference() const
		{
			return m_forwardReference;
		}

		/**
		 * Gets the name of the *commit* reference indicting pending commits from the external archive.
		 * Will return an empty string if there is no external archive configured (`isExternal()` returns`false`).
		 *
		 * @return the name of the *commit* reference.
		 */
		const std::string& getCommitReference() const
		{
			return m_commitReference;
		}
};


/**
 * Helper class to read and cache the content of the archive configuration.
 *
 * TODO: Though there is an `erad::imutils::Config` class, it is not thread safe.
 * A similar approach should be implemented there.
 */
class ArchiveConfigReader {
	private:
		/// An error message if the archive configuration could not be loaded
		std::string m_error;
		/// The path to the configuration file
		std::string m_configFile;

		/// Mutex for reloading the content
		std::mutex m_mutex;

		/// The current content of the configuration file
		std::shared_ptr<ArchiveConfig> m_config;

		/// the last modified time of the configuration file
		time_t lastModified;

		/// the time of next check
		time_t nextCheck;

		/// the default check interval, ie cache time (10 seconds)
		time_t configCheckInterval;

	public:
		/**
		 * Constructs a new Archive Config checker using the default config location (`$ARCHIVE_CONF`).
		 */
		ArchiveConfigReader();
		virtual ~ArchiveConfigReader();

		/**
		 * Gets the current configuration.
		 *
		 * Also checks if the configuration file has changed and reloads the configuration
		 * if needed. The most recent configuration will be returned.
		 *
		 * @return the current configuration
		 */
		std::shared_ptr<ArchiveConfig> getConfig();

		/**
		 * Checks if there was any error when loading the configuration.
		 *
		 * @return `true` if and only if there was an error
		 */
		bool hasError() const
		{
			return !m_error.empty();
		}

		/**
		 * Gets the error message if `hasError()` returns `true`.
		 *
		 * @return the error message (if any)
		 */
		const std::string& getError() const
		{
			return m_error;
		}
};


/**
 * Helper class to check if a study is deletable
 */
class DeleteChecker {
	private:
		/// Database connector
		erad::db::DBConnector& m_db;

		/// The current archive configuration
		std::shared_ptr<ArchiveConfig> m_archiveConfig;

		/// The study instance UID of the study to check
		std::string m_study;

		/**
		 * Checks the `STUDYDB` reference counter of the study.
		 *
		 * @return
		 * - `DeletableStatus::Objects` : Objects would remain, the STUDYDB reference would remain
		 * - `DeletableStatus::None`    : The STUDYDB reference counter would be gone
		 */
		DeletableStatus checkDb();

		/**
		 * Checks the `OBJECTFILE` reference counters of the study.
		 *
		 * @return
		 * - `DeletableStatus::Objects`    : Objects would remain after a delete
		 * - `DeletableStatus::ArchiveFwd` : Some object are not yet sent to the external archive (otherwise would be deletable)
		 * - `DeletableStatus::None`       : All objects would be gone
		 */
		DeletableStatus checkObjectFiles();

		/**
		 * Checks the `STUDYDIR` reference counter of the study.
		 *
		 * @return
		 * - `DeletableStatus::DIRECTORY` : The study directory would remain
		 * - `DeletableStatus::None`      : The study directory would be gone
		 */

		DeletableStatus checkDir();

	public:

		/**
		 * Constructs a `DeletableChecker` object to check if a study is deletable.
		 *
		 * @param db database connector
		 * @param archiveConfig the current archive configuration
		 * @param study the study instance UID of the study to check
		 */
		DeleteChecker( erad::db::DBConnector& db, std::shared_ptr<ArchiveConfig> archiveConfig, const std::string& study );

		/**
		 * Checks if all objects have been sent to the external archive, when an external archive is configured.
		 * That is checks whether all objects are sent (or at least queued) to the archive and the study can be safely deleted.
		 *
		 * Note, whether the entire study would be gone will not be checked.
		 *
		 * @return
		 * - `DeletableStatus::ArchiveFwd` : if any of the objects is not yet archived
		 * - `DeletableStatus::None`       : if all the objects are archived (or queued)
		 */
		DeletableStatus sentToArchive();

		/**
		 * Checks if the study directory would be entirely gone after a delete or part of the study would remain.
		 * Checks if the database entries, object files or the staudy directory is held by any references.
		 *
		 * @return
		 * - `DeletableStatus::Objects`    : Objects would remain after a delete
		 * - `DeletableStatus::ArchiveFwd` : Some object are not yet sent to the external archive (otherwise the object files would be deletable)
		 * - `DeletableStatus::DIRECTORY`  : The study directory would remain
		 * - `DeletableStatus::None`       : The entire study - including the study directory - would be gone
		 */
		DeletableStatus checkDeletable();
};


/**
 * A Helper class to quick delete a study.
 *
 * Quick delete drops all references from the `STUDYDIR`, `OBJECTFILE` and `STUDYDB` reference counters
 * without invoking the *on release* handler individually.
 * Invokes the *on release* handler of the study directory though, to cleanup the database entries and the directories.
 */
class QuickDeleter {
	private:
		/// Database connector
		erad::db::DBConnector& m_db;

	public:
		/**
		 * Constructs a `QuickDeleter` object
		 *
		 * @param db Database connector
		 */
		QuickDeleter( erad::db::DBConnector& db );

		/**
		 * Quick-deletes the provided study.
		 *
		 * Drops all the references (`STUDYDB`, `OBJECTFILE`, `STUDYDIR`) and cleans up the study directory
		 * and database entries by invoking the *on released* handler of the `STUDYDIR` reference counter.
		 *
		 * @param study the study to delete
		 *
		 * @return `true` if and only if the study was successfully deleted.
		 *
		 * @throw `std::exception` if an error occurs
		 */
		bool deleteStudy( const std::string& study );


		/**
		 * Quick-deletes the provided objects from a particular study.
		 *
		 * Drops all references from the object files without invokin the *on release* handler, and
		 * deletes the files from the study directory.
		 *
		 * Adjusts the *file* references on the `STUDYDIR` reference counter at the end.
		 *
		 * Note, that this method intended to be used by the Java study deleter tool, which will perform additional
		 * steps (for example updating the study and/or the report table, start the appropriate tasks, ...)
		 *
		 *
		 * @param study the study
		 * @param objects the objects to be deleted from the study
		 *
		 * @return `true` if and only if the objects were successfully deleted.
		 *
		 * @throw `std::exception` if an error occurs
		 */
		bool deleteObjects( const std::string& study, const std::list<std::string> objects );
};

}	// namespace erad::utils

#endif  // ERAD_UTILS_QUICKDELETER_H_